# -*- coding: utf8 -*-

from .Config import Config
from .Console import Console
from .Client import Client
from .Server import Server

from ..utils.utils import *
from ..utils.ColorString import ColorString
from ..core.Oracle import Oracle
from .Downloader import Downloader

class Game:

    def __init__(self, config):
        self.config = config
        self.console = Console(self.config)
        self.game = Client(self.config) if self.config.is_client else Server(self.config)
        
    def start(self):

        # 检查是否安装JDK
        Oracle.install_jdk()

        # 控制台用户交互展示
        self.console.userInteraction()

        try:
            #启动游戏
            self.game.start()
            
        except Exception as e:
            print(e)
            print(ColorString.warn('Start Failed!!!'))


# -*- coding: utf8 -*-

__all__ = [
    'OrzMC',
    'Game',
    'Client',
    'Server',
    'Config',
    'Console'
    'Constants',
    'Downloader',
    'Tools',
    'Version',
    'Nginx',
    'Daemon',
]


# -*- coding: utf8 -*-
import os
from .Config import Config
from ..utils.ColorString import ColorString

class Daemon:

    @classmethod
    def setup(cls, config):
        '''配置minecraft守护进程，使用systemctl管理'''

        daemon_name = 'Joker Minecraft Server'
        daemon_user = os.getlogin()
        daemon_mc_version = config.version
        daemon_mc_min_mem = config.mem_min
        daemon_mc_max_mem = config.mem_max
        daemon_mc_server_type = config.game_type
        daemon_mc_title = 'jokermc'
        orzmc_bin_path = os.popen('which orzmc').read().strip()
        screen_bin_path = os.popen('which screen').read().strip()
        sleep_bin_path = os.popen('which sleep').read().strip()
        daemon_mc_stop_secs = 10
        daemon_mc_restart = '60s'
        daemon_mc_work_dir = '~'
        systemctl_daemon_conf = f"""
[Unit]
Description={daemon_name}
After=network.target

[Service]
User={daemon_user}
WorkingDirectory={daemon_mc_work_dir}
Environment=TITLE='{daemon_mc_title}'
Environment=STOP_SHUTDOWN='服务器即将关闭, 正在保存地图...'
Environment=MC_VERSION='{daemon_mc_version}'
Environment=MC_MIN='{daemon_mc_min_mem}'
Environment=MC_MAX='{daemon_mc_max_mem}'
Environment=MC_SERVER_TYPE='{daemon_mc_server_type}'
ExecStart=/bin/sh -c '{screen_bin_path} -DmS $TITLE {orzmc_bin_path} -s -v ${{MC_VERSION}} -m ${{MC_MIN}} -x ${{MC_MAX}} -t ${{MC_SERVER_TYPE}}'
ExecReload={screen_bin_path} -p 0 -S $TITLE -X eval 'stuff "reload"\\015'
ExecStop={screen_bin_path} -p 0 -S $TITLE -X eval 'stuff "say ${{STOP_SHUTDOWN}}"\\015'
ExecStop={screen_bin_path} -p 0 -S $TITLE -X eval 'stuff "save-all"\\015'
ExecStop={screen_bin_path} -p 0 -S $TITLE -X eval 'stuff "stop"\\015'
ExecStop={sleep_bin_path} {daemon_mc_stop_secs}
Restart=on-failure
RestartSec={daemon_mc_restart}

[Install]
WantedBy=multi-user.target
"""
        try:
            systemctl_conf_file_path = Config.game_version_server_systemctl_conf_file_path()
            with open(systemctl_conf_file_path,'w',encoding='utf-8') as cfg:
                cfg.write(systemctl_daemon_conf)
                print(ColorString.confirm('minecraft.service has been writen in location: %s' % systemctl_conf_file_path))

            systemctl_system_dir = '/etc/systemd/system'
            minecraft_systemctl_conf_filename = os.path.basename(systemctl_conf_file_path)
            if os.path.exists(systemctl_system_dir) and os.path.isdir(systemctl_system_dir):
                minecraft_service_file_path = os.path.join(systemctl_system_dir, minecraft_systemctl_conf_filename)
                cmd = 'sudo ln -snf {source} {dest} && sudo systemctl daemon-reload && sudo systemctl enable {service} && sudo systemctl restart {service}'.format(
                    source = systemctl_conf_file_path,
                    dest = minecraft_service_file_path,
                    service = minecraft_systemctl_conf_filename
                )
                ret = os.system(cmd)
                if ret == 0:
                    print(ColorString.confirm('started service: %s' % minecraft_systemctl_conf_filename))
        except Exception as e:
            print(e)
            print(ColorString.error('minecraft daemon configuration for systemctl failed!'))

        



# -*- coding: utf8 -*-
import os
import json

class OptiFine:

    _configuration = None

    @classmethod
    def json_configuration(cls, config):
        if OptiFine._configuration == None:
            optifine_json_path = os.path.join(config.game_version_client_versions_dir(), config.lastVersionId, config.lastVersionId + '.json')
            if  'optifine' in config.lastVersionId.lower() and os.path.exists(optifine_json_path):
                with open(optifine_json_path, 'r') as f:
                    OptiFine._configuration = json.load(f)
        return OptiFine._configuration

    @classmethod
    def library_optifine_jar_paths(cls, config):
        ret = []
        optifine_library_dir = os.path.join(config.game_version_client_library_dir(), 'optifine')
        if 'optifine' in config.lastVersionId.lower() and os.path.exists(optifine_library_dir):
            for dirpath, _, filenames in os.walk(optifine_library_dir):
                for filename in filenames:
                    _, type = os.path.splitext(filename)
                    if type == '.jar':
                        path = os.path.join(dirpath, filename)
                        ret.append(path)
        return ret


# -*- coding: utf8 -*-
from .Config import Config
from ..utils.ColorString import ColorString
import os

class Nginx:
    @classmethod
    def setup(cls):
        '''配置并启动nginx'''
        nginx_config_file = Config.game_version_server_nginx_file_path()
        try:
            with open(nginx_config_file, 'w', encoding='utf-8') as cfg:
                cfg.write('\n'.join([
                    Nginx.web_file_server_conf(), 
                    Nginx.web_live_map_conf(), 
                    Nginx.web_skin_system_conf(),
                ]))
                print(ColorString.confirm('Nginx conf file location: %s' % nginx_config_file))
            
            # 创建nginx配置文件软链接
            nginx_conf_dir = '/etc/nginx/conf.d'
            if os.path.exists(nginx_conf_dir) and os.path.isdir(nginx_conf_dir):
                minecraft_nginx_conf_file = os.path.join(nginx_conf_dir, os.path.basename(nginx_config_file))
                cmd = 'sudo ln -snf %s %s' % (nginx_config_file, minecraft_nginx_conf_file)
                ret = os.system(cmd)
                if ret == 0:
                    print(ColorString.confirm('Create symbol link file: %s' % minecraft_nginx_conf_file))

                nginx_process_number = int(os.popen('ps -ef | grep nginx | grep -v grep | wc -l').read().strip())
                cmd = None
                if nginx_process_number > 0:
                    cmd = 'sudo nginx -s reload'
                else:
                    cmd = 'sudo nginx'

                if cmd:
                    ret = os.system(cmd)
                    if ret == 0:
                        print(ColorString.confirm('minecraft nginx config successfully!'))
                    else:
                        print(ColorString.error('minecraft nginx config failed!'))
                # 配置HTTPS
                Nginx.setupSSL()
        except Exception as e:
            print(e)
            print(ColorString.error('Config Nginx for Minecraft Failed!!!'))
            exit(-1)

    @classmethod
    def setupSSL(cls):
        '''配置nginx服务支持https'''

        cmd = 'sudo apt-get update ||'\
        'sudo apt-get install -y software-properties-common &&'\
        'sudo add-apt-repository -y universe &&'\
        'sudo add-apt-repository -y ppa:certbot/certbot &&'\
        'sudo apt-get update ||'\
        'sudo apt-get install -y certbot python3-certbot-nginx &&'\
        'sudo certbot --nginx'
        if os.system(cmd) == 0:
            print(ColorString.confirm('Config HTTPS successfully!'))
        else:
            print(ColorString.error("Config HTTPS failed!"))

    
    @classmethod
    def web_file_server_conf(cls):
        '''配置minecraft文件服务器'''
        port=80
        file_server_root_dir = Config.game_ftp_server_base_dir()
        server_domain = 'download.jokerhub.cn'

        return f"""
# 我的世界文件服务器
server {{
    listen {port};
    server_name {server_domain};
    root {file_server_root_dir};

    autoindex on;
    autoindex_exact_size off;
    autoindex_localtime on;

    location / {{
        try_files $uri $uri/ =404;
    }}
}}
"""

    @classmethod
    def web_skin_system_conf(cls):
        '''配置minecraft皮肤系统'''
        port=80
        file_server_root_dir = '/var/www/SkinSystem'
        server_domain = 'skin.jokerhub.cn'
        fastcgi_pass = 'unix:/run/php/php7.2-fpm.sock'
        return f"""
# Minecraft SkinSystem
server {{
    listen {port};
    server_name {server_domain};
    root {file_server_root_dir};

    # Add index.php to the list if you are using PHP
    index index.php index.html index.htm index.nginx-debian.html;
    location / {{
        try_files $uri $uri/ =404;
    }}

    # pass PHP scripts to FastCGI server
    location ~ \.php$ {{
        include snippets/fastcgi-php.conf;
        # With php-fpm (or other unix sockets):
        fastcgi_pass {fastcgi_pass};
    }}
}}
"""

    @classmethod
    def web_live_map_conf(cls):
        '''配置minecraft地图'''
        port=80
        server_domain = 'map.jokerhub.cn'
        return f"""
server {{
    listen {port};
    server_name {server_domain};
    location / {{
        proxy_pass      http://localhost:8123;
    }}
}}
"""



# -*- coding: utf8 -*-

from ..utils.ColorString import ColorString
from .Version import *

# 选择版本
ALL_VERSIONS_HINT = ColorString.warn('\nAll Release Versions as follow:\n')
TAB_SPACE = ''
LEADING_SPACE = '  '
VERSION_FORMATTER='{:10}'
SELECT_VERSION_HINT = ColorString.warn('\nPlease select a version number of above list to %s %s ')
DEFAULT_VERSION_HINT = ColorString.error('(default: %s):')
CHOOSED_VERSION = ColorString.confirm('You choose the version: %s')
NOT_FOUND_VERSION = ColorString.warn('There is no such a release version game, use default!')
CHOOSED_DEFAULT_VERSION = ColorString.confirm('You choose the default version(%s)!')

# 选择用户名
CHOOSE_USERNAME_HINT = ColorString.warn('Please choose a username %s ') % ColorString.error('(default: %s):')
CHOOSED_USERNAME = ColorString.confirm('You username in game is: %s')
CHOOSED_DEFAULT_USERNAME = ColorString.warn('Use the default username!!!')

# 工具版本号
ORZMC_VERSION_NUMBER_STR = ColorString.warn('OrzMC version: %s' % ORZMC_VERSION_NUMBER)

# BMCLAPI下载版权提示
BMCLAPI_DESC = ColorString.hint('\nCurrent file download service using ') + ColorString.confirm('BMCLAPI\n')


# -*- coding: utf8 -*-

# BMCLAPI文档地址: https://bmclapidoc.bangbang93.com

class BMCLAPI:
    mojang_bmclapi_domain_map = {
        'launchermeta.mojang.com': 'bmclapi2.bangbang93.com',
        'launcher.mojang.com': 'bmclapi2.bangbang93.com',
        'resources.download.minecraft.net': 'bmclapi2.bangbang93.com/assets',
        'libraries.minecraft.net': 'bmclapi2.bangbang93.com/maven',
    }


# -*- coding: utf8 -*-

class Plugins:
    _singleton = None
    @classmethod
    def singleton(cls, config):
        '''单例方法'''
        if not Plugins._singleton:
            Plugins._singleton = Plugins(config)
        return Plugins._singleton

    def __init__(self, config):
        self.config = config

    def login(self):
        '''登录相关插件'''
        pass

    def permission(self):
        '''权限分配控制插件'''
        pass

    def skinSystem(self):
        '''皮肤系统插件'''
        pass

    def transfer(self):
        '''传送插件'''
        pass
    
    def webWorld(self):
        '''地图web浏览'''
        pass

    def recovery(self):
        '''恢复插件'''
        pass


# -*- coding: utf8 -*-
from bs4 import BeautifulSoup
import os, requests
from ..utils.ColorString import ColorString

class Forge:

    def __init__(self, version):
        self.version = version
        self.fullVersion = None
        self.briefVersion = None
        self.version_homePage = 'https://files.minecraftforge.net/maven/net/minecraftforge/forge/index_%s.html' % self.version
        print(ColorString.warn('parsing the forge installer download link...'))
        html = requests.get(self.version_homePage).text
        soup = BeautifulSoup(html, 'html.parser')
        installer = list(map(lambda i: i.parent.get('href'), soup.select('.classifier-installer')))[0]
        if len(installer) > 0:
            self.forge_installer_url = installer.split('&')[1].split('=')[1]
            self.fullVersion = '-'.join(os.path.basename(self.forge_installer_url).split('-')[0:-1])
            self.briefVersion = '-'.join(self.fullVersion.split('-')[0::2])
            print(ColorString.confirm('Get the forge installer download link!!'))
        else:
            print(ColorString.error( 'Not Found Any Installer!!!!'))

            




# -*- coding: utf8 -*-
# papermc: https://papermc.io
# api v1: https://paper.readthedocs.io/en/latest/site/api.html
# api v2: https://papermc.io/api/docs/swagger-ui/index.html?configUrl=/api/openapi/swagger-config

import urllib.request
import urllib.parse
import json

class PaperAPI:
    ''' api documentation: https://paper.readthedocs.io/en/stable/site/api.html '''
    API = 'https://papermc.io/api/%(API_VERSION)s/%(PROJECT_NAME)s/%(PROJECT_VERSION)s/%(BUILD_ID)s/download'
    @classmethod
    def downloadURLV1(cls, project_name = 'paper', project_version = None, build_id = 'latest'):
        return   PaperAPI.API % {
            'API_VERSION': 'v1',
            'PROJECT_NAME': project_name,
            'PROJECT_VERSION': project_version,
            'BUILD_ID': build_id   
        }

    BASE_URL = 'https://papermc.io'
    @classmethod
    def downloadURLV2(cls, version):
        url = 'https://papermc.io/api/'
        url +='v2/'
        url += 'projects/paper/'
        url += 'versions/%s' % version
        jsonResp = json.loads(urllib.request.urlopen(url).read().decode('utf-8'))
        builds = jsonResp.get('builds')
        latest_build = max(builds)

        url += '/builds/%s' % latest_build
        jsonResp = json.loads(urllib.request.urlopen(url).read().decode('utf-8'))
        jar_name = jsonResp.get('downloads').get('application').get('name')

        url += '/downloads/%s' % jar_name
        return url






# -*- coding: utf8 -*-

from os.path import isfile
from .Config import Config
from .Downloader import Downloader
from .Nginx import Nginx
from .Daemon import Daemon
from ..utils.ColorString import ColorString
from ..utils.CleanUp import CleanUp
from ..utils.utils import *

from ..core.Spigot import Spigot

import os
import io
import time
import shutil 

class Server:
    def __init__(self, config):
        self.config = config
        self.downloader = Downloader(self.config)
        
    def start(self):
        '''start minecraft server'''    
        if self.config.is_client:
            return

        is_return = False

        # 配置Nginx服务
        if self.config.nginx:
            Nginx.setup()
            is_return = True

        if self.config.deamon:
            Daemon.setup(self.config)
            is_return = True

        if self.config.backup:
            self.backupWorld()
            is_return = True

        if is_return:
            return
        
        if self.config.isPure:
            self.downloader.downloadGameJSON()
            self.downloader.downloadServer()
            (serverJARFilePath, _, _) = self.serverJARFilePath()
            jarFilePath = serverJARFilePath
        elif self.config.isSpigot:
            if self.config.force_download or not os.path.exists(self.config.game_version_server_jar_file_path()):
                self.buildSpigotServer()
            jarFilePath = self.config.game_version_server_jar_file_path()
        elif self.config.isPaper:
            if self.config.force_download or not os.path.exists(self.config.game_version_server_jar_file_path()):
                self.downloader.downloadPaperServerJarFile()
            jarFilePath = self.config.game_version_server_jar_file_path()
        elif self.config.isForge:
            self.config.getForgeInfo()
            if self.config.force_download or not os.path.exists(self.config.game_version_server_jar_file_path()):
                self.buildForgeServer()
            jarFilePath = self.config.game_version_server_jar_file_path()
            if not os.path.exists(jarFilePath):
                jarFilePath = jarFilePath.replace(self.config.forgeInfo.fullVersion, self.config.forgeInfo.fullVersion + '-universal')
        else:
            print(ColorString.warn('Your choosed server is not exist!!!\nCurrently, there are three type server: pure/spigot/forge'))
            return

        jvm_opts = ' '.join([
            '-server',
            '-Xms' + self.config.mem_min,
            '-Xmx' + self.config.mem_max
        ])
        jarArgs = ['--forceUpgrade', 'nogui'] if (self.config.isSpigot or self.config.isPaper) and self.config.force_upgrade else ['nogui']
        self.startServer(self.startCommand(jvm_opts= jvm_opts, serverJARFilePath = jarFilePath, jarArgs = jarArgs))
        print(ColorString.confirm('Start Server Successfully!!!'))
    
    def serverJARFilePath(self):
        '''服务端运行需要的JAR文件所在路径'''
        server = self.config.game_version_json_obj().get('downloads').get('server')
        serverUrl = server.get('url')
        sha1 = server.get('sha1')
        serverJARFilePath = self.config.game_version_server_jar_file_path()
        return (serverJARFilePath, serverUrl, sha1)
    
    def downloadServer(self):    
        '''Download Server Jar File'''
        (serverJARFilePath, serverUrl, sha1) = self.serverJARFilePath()
        if not checkFileExist(serverJARFilePath, sha1):
            self.downloader.download(
                serverUrl,
                self.config.game_version_server_dir(), 
                self.config.game_version_server_jar_filename(), 
                prefix_desc='server jar file'
            )
        else:
            print("server jar file existed!")

    def startCommand(self, jvm_opts = '', serverJARFilePath = '', jarArgs = ['nogui']):
        '''construct server start command'''
        argList = [
            'java',
            jvm_opts,
            '-jar',
            serverJARFilePath
        ]
        argList.extend(jarArgs)
        cmd = ' '.join(argList)
        return cmd

    def checkEULA(self):
        return os.path.exists(self.config.game_version_server_eula_file_path())

    def startServer(self, cmd):
        '''启动minecraft服务器'''

        os.chdir(self.config.game_version_server_dir())

        # 如果没有eula.txt文件，则启动服务器生成
        if not self.checkEULA():
            if self.config.isForge:
                self.generateForgeServerEULA()
            else:
                os.system(cmd)

        # 同意eula
        with io.open(self.config.game_version_server_eula_file_path(), 'r', encoding = 'utf-8') as f:
            eula = f.read()
            checkEULA = eula.replace('false', 'true')
        with io.open(self.config.game_version_server_eula_file_path(), 'w', encoding = 'utf-8') as f:
            f.write(checkEULA)
        
        # 启动服务器
        
        if self.config.debug:
            print(cmd)

        os.system(cmd)

        # 设置服务器属性为离线模式
        with io.open(self.config.game_version_server_properties_file_path(), 'r', encoding = 'utf-8') as f:
            properties = f.read()
            if 'online-mode=false' in properties:
                offline_properties = None
            else:
                offline_properties = properties.replace('online-mode=true', 'online-mode=false')

        if offline_properties != None:
            with io.open(self.config.game_version_server_properties_file_path(), 'w', encoding = 'utf-8') as f:
                f.write(offline_properties)
                print(ColorString.confirm('Setting the server to offline mode, next launch this setting take effect!!!'))

        # 为服务器核心数据文件创建符号链接
        self.symlink_server_core_files_if_need()


    # 构建SpigotServer
    def buildSpigotServer(self):
        '''构建SpigotServer'''

        version = self.config.version
        spigot = Spigot(version)
        self.download(
            spigot.build_tool_jar, 
            self.config.game_version_server_build_dir(), 
            prefix_desc='spigot build tool jar file'
        )
        buildToolJarName = os.path.basename(spigot.build_tool_jar)
        buildToolJarPath = os.path.join(self.config.game_version_server_build_dir(), buildToolJarName)
        versionCmd = ' --rev ' + version if len(version) > 0 else 'lastest'
        linuxCmd = 'git config --global --unset core.autocrlf; java -jar ' + buildToolJarPath + versionCmd
        macCmd = 'export MAVEN_OPTS="-Xmx2G"; java -Xmx2G -jar ' + buildToolJarPath + versionCmd
        windowCmd = 'java -jar ' + buildToolJarPath + versionCmd
        cmd = macCmd if platformType() == 'osx' else linuxCmd if platformType() == 'linux' else windowCmd
        print(ColorString.warn('Start build the server jar file with build tool...'))
        os.chdir(self.config.game_version_server_build_dir())
        os.system(cmd)
        print(ColorString.confirm('Completed! And the spigot built server file generated!'))
        shutil.move(self.config.game_version_server_jar_file_path(isInBuildDir=True), self.config.game_version_server_jar_file_path())
        os.chdir(self.config.game_version_server_dir())
        shutil.rmtree(self.config.game_version_server_build_dir())
    
    # 构建Forge服务器
    def buildForgeServer(self):
        '''构建Forge服务器'''

        self.download(
            self.config.forgeInfo.forge_installer_url, 
            self.config.game_version_server_dir(), 
            prefix_desc='forge installer jar file'
        )

        installerJarFilePath = os.path.basename(self.config.forgeInfo.forge_installer_url)
        installServerCmd = 'java -jar ' + installerJarFilePath + ' --installServer'
        print(ColorString.warn('Start install the forge server jar file ...'))
        os.chdir(self.config.game_version_server_dir())
        os.system(installServerCmd)
        print(ColorString.confirm('Completed! And the forge server file generated!'))


    def generateForgeServerEULA(self):
        if self.config.isForge and not self.checkEULA(): 
            pure_server_jar = os.path.join(self.config.game_version_server_dir(), '.'.join(['minecraft_server',self.config.version,'jar']))
            boot_cmd = 'java -jar ' + pure_server_jar
            os.system(boot_cmd)

    def backupWorld(self):
        world_paths = self.config.game_version_server_world_dirs()
        if world_paths:            
            backup_path = self.config.game_version_server_world_backup_dir()
            now = time.localtime()
            fileName = '_'.join([time.strftime('%Y-%m-%dT%H:%M:%S', now), self.config.game_type, self.config.version]) + '.zip'
            world_backup_file = os.path.join(backup_path, fileName)

            def backupWorld_cleanUp():
                if os.path.isfile(world_backup_file) and os.path.exists(world_backup_file):
                    os.remove(world_backup_file)
                    print(ColorString.warn("Removed Invalid World Backup File: %s" % world_backup_file))

            print(ColorString.hint('Start Executing ZIP ...'))
            CleanUp.registerTask('backupWorld_cleanUp', backupWorld_cleanUp)
            zip(world_paths, world_backup_file)
            CleanUp.cancelTask('backupWorld_cleanUp')
            file_size = os.path.getsize(world_backup_file) / 1024.0 / 1024.0 / 1024.0
            print(ColorString.confirm("Completed! backuped world file(%.2fG): %s!!!" % (file_size, world_backup_file)))
        else: 
            print(ColorString.error('There is no world directory!!!'))

    def symlink_server_core_files_if_need(self):
        '''为服务器核心文件创建符号链接到共享目录，方便版本升级'''
        if self.config.symlink:
            # 软链接需要涉及的文件和目录
            destinations = [
                self.config.game_version_server_eula_file_path(),
                self.config.game_version_server_properties_file_path(),
                self.config.game_version_server_icon_file_path(),
                self.config.game_version_server_plugin_dir(),
            ]
            world_paths = self.config.game_version_server_world_dirs()
            if world_paths and len(world_paths) > 0:
                destinations.extend(self.config.game_version_server_world_dirs())

            # 遍历创建软链接
            for destination in destinations:
                # 软链接需要链接的源文件或源目录
                source = os.path.join(self.config.game_version_server_symlink_source_dir(),os.path.basename(destination))
                # 如果软链接生成的目录下有同名的文件和目录，则取个备份用的名称
                destination_backup = destination+'.before_symlink'

                if not os.path.exists(source):
                    if os.path.exists(destination):
                        shutil.move(destination, source)
                    else:
                        # 之前有备份过的同名文件和目录，则恢复同名的文件和目录
                        if os.path.exists(destination_backup):
                            shutil.move(destination_backup, destination)
                        continue

                # 如果已经创建了软链接，则跳过不再创建
                if os.path.islink(destination):
                    target = os.readlink(destination)
                    if os.path.normcase(target) == os.path.normcase(source):
                        continue
                
                # 如果软链接生成的目录下有同名文件或目录，则先进行备份
                if os.path.exists(destination):
                    shutil.move(destination, destination_backup)
                
                # 生成软链接
                os.symlink(source, destination, target_is_directory=os.path.isdir(source))
                print("%s -> %s" % (source, destination))


# -*- coding: utf8 -*-

from .utils import platformType

from enum import IntEnum

class BGColor(IntEnum):
    BLACK = 40
    RED = 41
    GREEN = 42
    YELLOW = 43
    BLUE = 44
    PURPLE = 45
    CYAN = 46
    WHITE = 47
    DEFAULT = 0
class FGColor(IntEnum):
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    PURPLE = 35
    CYAN = 36
    WHITE = 37
class DisplayMode(IntEnum):
    DEFUALT = 0
    HIGHLIGHT = 1
    UNDERLINE = 2
    TWINKLE = 3
    INVERSE = 4
    INVISIBLE = 5

class ColorString:

    @classmethod
    def string(cls, str, fg = FGColor.WHITE, bg = BGColor.DEFAULT, displayMode = DisplayMode.DEFUALT):
        if bg == BGColor.DEFAULT:
            return '\033[%s;%sm%s\033[0m' % (displayMode.value, fg.value, str) if platformType() != 'windows' else str
        else:
            return '\033[%s;%s;%sm%s\033[0m' % (displayMode.value, fg.value, bg.value, str) if platformType() != 'windows' else str

    @classmethod
    def warn(cls, str):
        return ColorString.string(str, FGColor.YELLOW, displayMode=DisplayMode.HIGHLIGHT)

    @classmethod
    def confirm(cls, str):
        return ColorString.string(str, FGColor.GREEN, displayMode=DisplayMode.HIGHLIGHT)

    @classmethod
    def error(cls, str):
        return ColorString.string(str, FGColor.RED, displayMode=DisplayMode.HIGHLIGHT)

    @classmethod
    def hint(cls, str):
        return ColorString.string(str, FGColor.YELLOW, displayMode=DisplayMode.HIGHLIGHT)


# -*- coding: utf8 -*-
from ..utils.utils import platformType
from ..utils.ColorString import ColorString
import requests
import os
from tqdm import tqdm

class Oracle:

    ''' 
    # Example Download Link
    ## Windows： 
    - https://download.oracle.com/otn-pub/java/jdk/12.0.2+10/e482c34c86bd4bf8b56c0b35558996b9/jdk-12.0.2_windows-x64_bin.exe

    ## MacOS:
    - https://download.oracle.com/otn-pub/java/jdk/12.0.2+10/e482c34c86bd4bf8b56c0b35558996b9/jdk-12.0.2_osx-x64_bin.dmg

    ## Ubuntu:
     https://download.oracle.com/otn-pub/java/jdk/12.0.2+10/e482c34c86bd4bf8b56c0b35558996b9/jdk-12.0.2_linux-x64_bin.deb
    '''

    VERSION_LATEST = ('12.0.2','10', 'e482c34c86bd4bf8b56c0b35558996b9')

    JDK_NAME = {
        'windows': 'windows-x64_bin.exe',
        'osx': 'osx-x64_bin.dmg',
        'linux': 'linux-x64_bin.deb'
    }

    @classmethod
    def install_jdk(cls, version = VERSION_LATEST):

        if not Oracle.check_jdk_installed():
            try:
                Oracle.download_jdk()  
                if not Oracle.execute_install_jdk():
                    print(ColorString.warn('Install JDK Failed!'))
                    exit(0)
                else:
                    input(ColorString.hint('Press Enter to continue...'))
            except Exception as e:
                print(e)
            finally:
                if Oracle.check_jdk_installed():
                    os.remove(Oracle.jdk_temp_filename())
        else:
            print('\n')

    @classmethod
    def check_jdk_installed(cls):
        return os.system('java -version') == 0

    @classmethod
    def download_jdk(cls):
        headers = {
            'user-agent': 'OrzMC'
        }

        cookies = {
            'oraclelicense': 'accept-securebackup-cookie'
        }
        requests.packages.urllib3.disable_warnings()
        response = requests.get(Oracle.jdk_url(), stream=True, verify=False, allow_redirects=True, cookies = cookies, headers = headers)
        total_length = int(response.headers.get('content-length'))
        mb_chunk = 1024 * 1024
        mb_size = int(total_length / mb_chunk + 0.5)
        with open(Oracle.jdk_temp_filename(), 'wb') as f:
            desc = os.path.basename(Oracle.jdk_url())
            for chunk in tqdm(iterable=response.iter_content(mb_chunk),total=mb_size,unit='MB',desc=desc):
                f.write(chunk)

    @classmethod
    def jdk_url(cls):
        (version, build, hash) = Oracle.VERSION_LATEST
        return 'https://download.oracle.com/otn-pub/java/jdk/%s+%s/%s/jdk-%s_%s' % (version, build, hash, version, Oracle.jdk_name())

    @classmethod
    def jdk_name(cls):
        return Oracle.JDK_NAME.get(platformType())

    @classmethod
    def jdk_temp_filename(cls):
        return 'jdk-' + Oracle.jdk_name()

    @classmethod
    def execute_install_jdk(cls):
        return Oracle.execute_commands(Oracle.commands().get(platformType()).get('install'))

    @classmethod
    def execute_uninstall_jdk(cls):
        return Oracle.execute_commands(Oracle.commands().get(platformType()).get('uninstall'))

    @classmethod
    def execute_commands(cls, cmds, hint = ColorString.hint('May be you should input the user password to continue')):
        print(hint)
        if isinstance(cmds, list):
            ret = True
            for cmd in cmds:
                cur_ret =(os.system(cmd) == 0)
                if not cur_ret:
                    print('cmd: %s failed' % cmd)
                    ret = False
                    break
            return ret
        else:
            return False

    @classmethod
    def commands(cls):
        return  {
            'windows': {
                'install': [
                    
                ],
                'uninstall': [

                ]
            },
            'osx': {
                'install': [
                    'open ' + Oracle.jdk_temp_filename(),
                ],
                'uninstall': [
                    'sudo rm -R /Library/Java/JavaVirtualMachines/jdk-%s.jdk' % Oracle.VERSION_LATEST[0],
                    'sudo pkgutil --forget com.oracle.jdk-%s' % Oracle.VERSION_LATEST[0]
                ]
            },
            'linux': {
                'install': [
                ],
                'uninstall': [

                ]
            }
        }


# -*- coding: utf8 -*-

import os, json
from ..utils.utils import *
from ..core.Forge import Forge

class Config:
    '''Public Definitions'''
    GAME_DEFAULT_USERNAME = "guest"
    GAME_TYPE_PURE = 'vanilla'
    GAME_TYPE_SPIGOT = 'spigot'
    GAME_TYPE_FORGE = 'forge'
    GAME_TYPE_PAPER = 'paper'
    
    '''Private Definitions'''
    BASE_PATH = os.path.expanduser('~')
    GAME_ROOT_DIR = os.path.join(BASE_PATH,'minecraft')
    GAME_VERSIONS_DIR = os.path.join(GAME_ROOT_DIR,'versions')


    def __init__(self,args):

        # 调试信息
        if args.debug:
            print(args)

        self.is_client = not args.server
        self.version = args.version
        self.username = args.username
        self.mem_min = args.minmem
        self.mem_max = args.maxmem
        self.is_extract_music = args.extract_music
        self.symlink = args.symlink
        self.bmclapi = args.bmclapi
        self.nginx = args.nginx
        self.deamon = args.deamon

        game_type = args.type
        self.isPure = (game_type == Config.GAME_TYPE_PURE)
        self.isSpigot = (game_type == Config.GAME_TYPE_SPIGOT)
        self.isForge = (game_type == Config.GAME_TYPE_FORGE)
        self.isPaper = (game_type == Config.GAME_TYPE_PAPER)
        if self.isPure:
            self.game_type = Config.GAME_TYPE_PURE
        elif self.isSpigot:
            self.game_type = Config.GAME_TYPE_SPIGOT
        elif self.isPaper:
            self.game_type = Config.GAME_TYPE_PAPER
        elif self.isForge: 
            self.game_type = Config.GAME_TYPE_FORGE
        else:
            self.game_type = ''

        self.debug = args.debug
        self.force_upgrade = args.force_upgrade_world
        self.backup = args.backup_world
        self.optifine = args.optifine
        self.api = args.api
        self.force_download = args.force_download

    def getForgeInfo(self):
        if self.isForge:
            self.forgeInfo = Forge(version = self.version)
        else:
            self.forgeInfo = None

    def java_class_path_list_separator(self):
        return ';' if platformType() == 'windows' else ':'

    @classmethod
    def game_root_dir(cls):
        '''Game Root Dir'''
        makedirs(Config.GAME_ROOT_DIR)
        return Config.GAME_ROOT_DIR

    def game_versions_dir(self):
        '''Game Versions Dir'''
        makedirs(Config.GAME_VERSIONS_DIR)
        return Config.GAME_VERSIONS_DIR

    def game_version_dir(self):
        '''Version Related Directory'''
        version_dir = os.path.join(self.game_versions_dir(),self.version)
        makedirs(version_dir)
        return version_dir

    ### Client
    def game_version_client_dir(self):
        client_dir = os.path.join(self.game_version_dir(), 'client', Config.GAME_TYPE_PURE)
        makedirs(client_dir)
        return client_dir

    def game_version_client_assets_dir(self):
        assets_root_dir = os.path.join(self.game_version_client_dir(), 'assets')
        makedirs(assets_root_dir)
        return assets_root_dir

    def game_version_client_assets_indexs_dir(self):
        assets_indexs_dir = os.path.join(self.game_version_client_assets_dir(), 'indexes')
        makedirs(assets_indexs_dir)
        return assets_indexs_dir

    def game_version_client_assets_objects_dir(self, hash):
        assets_objects_dir = os.path.join(self.game_version_client_assets_dir(), 'objects', hash[0:2])
        makedirs(assets_objects_dir)
        return assets_objects_dir

    def game_version_client_library_dir(self, subpath = None):
        lib_dir = os.path.join(self.game_version_client_dir(), 'libraries')
        if None != subpath:
            subdir =  os.path.dirname(subpath)
            lib_dir = os.path.join(lib_dir,subdir)
        makedirs(lib_dir)
        return lib_dir
    
    def game_version_client_versions_dir(self):
        dir = os.path.join(self.game_version_client_dir(),'versions')
        makedirs(dir)
        return dir

    def game_version_client_versions_version_dir(self):
        dir = os.path.join(self.game_version_client_versions_dir(), self.version)
        makedirs(dir)
        return dir

    def game_version_client_native_library_dir(self):
        native_lib_dir = os.path.join(self.game_version_client_versions_version_dir(), self.version + '-natives')
        makedirs(native_lib_dir)
        return native_lib_dir

    def game_version_json_file_path(self):
        return os.path.join(self.game_version_client_versions_version_dir(), self.version + '.json')

    def game_version_client_jar_filename(self):
        return self.version + '.jar'

    def game_version_client_jar_file_path(self):
        jar_file_path = os.path.join(self.game_version_client_versions_version_dir(), self.game_version_client_jar_filename())
        return jar_file_path

    def game_version_forge_json_file_path(self):
        forge_json_file_name = '-'.join([self.version, self.forgeInfo.briefVersion]) + '.json'
        return os.path.join(self.game_version_client_dir(), forge_json_file_name)
    
    def game_version_launcher_profiles_json_path(self):
        return os.path.join(self.game_version_client_dir(), 'launcher_profiles.json')

    def game_version_json_obj(self):
        return loadJSON(self.game_version_forge_json_file_path() if self.isForge else self.game_version_json_file_path())

    def game_version_json_assets_obj(self):
        index_json_url = self.game_version_json_obj().get('assetIndex').get('url')
        index_json_path = os.path.join(self.game_version_client_assets_indexs_dir(), os.path.basename(index_json_url))
        return loadJSON(index_json_path)

    ### Server
    def game_version_server_dir(self):
        server_dir = os.path.join(self.game_version_dir(), 'server', self.game_type)
        makedirs(server_dir)
        return server_dir

    def game_version_server_jar_filename(self):
        if self.isForge:
            return self.forgeInfo.fullVersion + '.jar'
        elif self.isSpigot:
            return 'spigot-' + self.version + '.jar'
        elif self.isPaper:
            return 'paper-' + self.version + '.jar'
        elif self.isPure:
            return self.version + '.jar'
        else:
            return ''

    def game_version_server_jar_file_path(self, isInBuildDir=False):
        jar_file_path = os.path.join(self.game_version_server_build_dir() if isInBuildDir else self.game_version_server_dir(), self.game_version_server_jar_filename())
        return jar_file_path

    def game_version_server_eula_file_path(self):
        return os.path.join(self.game_version_server_dir(), 'eula.txt')

    def game_version_server_properties_file_path(self):
        return os.path.join(self.game_version_server_dir(), 'server.properties')

    def game_version_server_build_dir(self):
        build_path = os.path.join(self.game_version_server_dir(), 'build')
        makedirs(build_path)
        return build_path
    
    def game_version_server_world_dirs(self):
        worldName = 'world'
        properties_file_path = self.game_version_server_properties_file_path()

        if os.path.exists(properties_file_path):
            with open(properties_file_path, 'r') as f:
                for line in f.readlines():
                    if 'level-name' in line:
                        worldName = line.strip('\n').split('=')[1]
        else:
            return None

        game_dir = self.game_version_server_dir()

        world_dirs = []
        world_dir = os.path.join(game_dir,worldName)
        if world_dir and os.path.exists(world_dir):
            world_dirs.append(world_dir)

        if self.isSpigot or self.isPaper:
            world_nether_dir = world_dir + '_nether'
            world_the_end_dir = world_dir + '_the_end'
            if world_nether_dir and os.path.exists(world_nether_dir):
                world_dirs.append(world_nether_dir)
            if world_the_end_dir and os.path.exists(world_the_end_dir):
                world_dirs.append(world_the_end_dir)

        if len(world_dirs):
            return world_dirs
        else:
            return None

    def game_version_server_icon_file_path(self):
        return os.path.join(self.game_version_server_dir(), 'server-icon.png')

    def game_version_server_plugin_dir(self):
        return os.path.join(self.game_version_server_dir(),'plugins') if self.isPaper else None

    def game_version_server_world_backup_dir(self):
        game_world_backup_dir = os.path.join(Config.game_ftp_server_base_dir(), 'game_backup')
        makedirs(game_world_backup_dir)
        return game_world_backup_dir
    
    def game_version_server_symlink_source_dir(self):
        symlink_destination_dir = Config.game_ftp_server_core_data_backup_dir()
        makedirs(symlink_destination_dir)
        return symlink_destination_dir

    def game_version_client_mp3_dir(self):
        mp3_dir = os.path.join(Config.game_ftp_server_base_dir(),'client_music', self.version)
        makedirs(mp3_dir)
        return mp3_dir
        
    def game_download_temp_dir(self):
        download_temp_dir = os.path.join(Config.GAME_ROOT_DIR, 'download_tmp_dir')
        makedirs(download_temp_dir)
        return download_temp_dir

    @classmethod
    def game_ftp_server_base_dir(cls):
        ftp_base_dir = os.path.join(Config.BASE_PATH, 'minecraft_world_backup')
        makedirs(ftp_base_dir)
        return ftp_base_dir
        
    @classmethod
    def game_ftp_server_core_data_backup_dir(cls):
        server_core_data_backup_dir = os.path.join(Config.game_ftp_server_base_dir(),'mcserver')
        makedirs(server_core_data_backup_dir)
        return server_core_data_backup_dir

    @classmethod
    def game_config_dir(cls):
        game_config_dir = os.path.join(Config.GAME_ROOT_DIR, 'configurations')
        makedirs(game_config_dir)
        return game_config_dir

    @classmethod
    def game_version_server_nginx_file_path(cls):
        nginx_conf_file_path = os.path.join(Config.game_config_dir(),'nginx_minecraft.conf')
        return nginx_conf_file_path

    @classmethod
    def game_version_server_systemctl_conf_file_path(cls):
        systemctl_conf_file_path = os.path.join(Config.game_config_dir(),'minecraft.service')
        return systemctl_conf_file_path


# -*- coding: utf8 -*-

class CleanUp:

    is_sigint_up = False
    cleanTasks = {}

    @classmethod
    def registerTask(cls, key, value):
        CleanUp.cleanTasks[key] = value

    @classmethod
    def cancelTask(cls, key):
        if key in CleanUp.cleanTasks.keys():
            CleanUp.cleanTasks.pop(key)
    
    @classmethod
    def executeCleanTask(cls, key = None):
        if key == None: 
            for task in CleanUp.cleanTasks.values():
                task()
        else:
            if key in CleanUp.cleanTasks.keys():
                CleanUp.cleanTasks[key]() 
        
    @classmethod
    def sigint_handler(cls, signum, frame):
        CleanUp.is_sigint_up = True
        print(ColorString.warn("\nForce Exit!"))
        CleanUp.executeCleanTask()
        exit(-1)


import signal
from .utils import platformType
from .ColorString import ColorString
signal.signal(signal.SIGINT, CleanUp.sigint_handler)
signal.signal(signal.SIGTERM, CleanUp.sigint_handler)
if platformType() != 'windows':
    signal.signal(signal.SIGHUP, CleanUp.sigint_handler)



# -*- coding: utf8 -*-
from .Config import Config
from ..core.Mojang import Mojang
from .Constants import *
from ..utils.utils import hint

import os
import json

class Console:
    
    def __init__(self, config):
        '''初始化控制台实例'''
        self.config = config
        
    def userInteraction(self):
        '''控制台展示用户交互界面'''
        if self.config.version == None:
            # 如果没有提示版本参数，则展示可选版本号供用户选择
            self.showVersionList()

        if self.config.username == Config.GAME_DEFAULT_USERNAME and not self.config.is_extract_music:
            # 如果用户名为默认值时，提示用户选择一个自己的用户名进行游戏
            self.showUserName()

        # 客户端启动方式
        self.selectLauncherProfile()

    def showVersionList(self):
        '''控制台交互显示可选客户端版本号'''
        
        # 显示所有版本提示
        print(ALL_VERSIONS_HINT)
        releaseVersions = Mojang.get_release_version_id_list(update = True)

        versionInfo = LEADING_SPACE
        for index, releaseVersion in enumerate(releaseVersions): 
            versionInfo += VERSION_FORMATTER.format(str(releaseVersion))
            if (index + 1) % 5 == 0:
                versionInfo += '\n' + LEADING_SPACE
            else:
                versionInfo += TAB_SPACE
            if releaseVersion == '1.13': 
                break

        print(ColorString.confirm(versionInfo))

        if len(releaseVersions) > 0:
            self.config.version = releaseVersions[0] # 默认版本号

        select = hint(SELECT_VERSION_HINT % ('deploy' if not self.config.is_client else 'play' , DEFAULT_VERSION_HINT % self.config.version))

        if len(select) > 0:
            found = False
            for releaseVersion in releaseVersions: 
                if releaseVersion == select.strip():
                    found = True
                    self.config.version = releaseVersion
                    print(CHOOSED_VERSION % self.config.version)
            if not found:
                print(NOT_FOUND_VERSION)
        else:
            print(CHOOSED_DEFAULT_VERSION % self.config.version)

        print('\n')
    
    def showUserName(self):
        '''控制台交互输入玩家用户名'''
        
        # 只针对客户端有效
        if not self.config.is_client:
            return 

        u = hint(CHOOSE_USERNAME_HINT % self.config.username)

        if len(u) > 0:
            self.config.username = u
            print(CHOOSED_USERNAME % self.config.username)
        else:
            print(CHOOSED_DEFAULT_USERNAME)

        print('\n')


    def selectLauncherProfile(self):

        if not self.config.is_client or not self.config.isPure or not self.config.optifine:
            return 

        launcher_profiles_json_file_path = self.config.game_version_launcher_profiles_json_path()
        if os.path.exists(launcher_profiles_json_file_path):
            content = None
            with open(launcher_profiles_json_file_path, 'r') as f:
                content = json.load(f)
                profiles = content.get('profiles', None)
                count = len(profiles)
                if count > 0:
                    print(ColorString.warn("There are those profiles you can choose to launch: "))
                    index = 0
                    selectedIndex = 0
                    keys = profiles.keys()
                    for key in keys:
                        index = index + 1
                        if key == content['selectedProfile']:
                            selectedIndex = index
                        isSelected = '*' if key == content['selectedProfile'] else ' '
                        print(ColorString.hint('\t%s %s. %s' % (isSelected, index, key)))

                    options = "[1 - %s] " % count if count > 1 else ""
                    try:
                        input = hint(ColorString.warn("Which one you choose %s: " % options))
                        which = int(input) if input and len(input) > 0 else selectedIndex
                        if which >= 1 and which <= count:
                            selected_key = list(keys)[which - 1]
                            self.config.lastVersionId = profiles.get(selected_key).get('lastVersionId')
                            content['selectedProfile'] = selected_key
                        else: 
                            print(ColorString.error("There is no option you specified!"))
                            exit(-1)
                    except:
                        print(ColorString.error("There is no option you specified!"))
                        exit(-1)

            if content != None: 
                with open(launcher_profiles_json_file_path, 'w') as f: 
                    json.dump(content, f)